import numpy as np

class Track():

  def __init__(self, unique_id, parent_id, pdg_id):
    self.unique_id = unique_id
    self.parent_id = parent_id
    self.pdg_id = pdg_id
    self.points = []

  def __str__(self):
    return f"UID: {self.unique_id},  parent_ID: {self.parent_id},  PDGID: {self.pdg_id}\n" + \
           f"points:\n" + \
           f"  {self.points[0]}\n" + \
           f"  ... \n" + \
           f"  {self.points[-1]}\n"

  def __repr__(self):
    return self.__str__()

  def AddPoint(self, point):
    self.points.append(point)
    pass



class DataBuffer():

  def __init__(self):
    self.vertices = []
    self.tracks = {}
    self.photon_cloud = []
    self.status = 'empty' # 'empty' -> 'in_progress' -> 'ready'

  def AddVertex(self, vertex):
    self.status = 'in_progress'
    self.vertices.append(vertex)

  def AddPhoton(self, photon):
    self.status = 'in_progress'
    self.photon_cloud.append(photon)

  def CreateTrack(self, uid, parent_id, pdg_id):
    self.status = 'in_progress'
    self.tracks[uid] = Track(uid, parent_id, pdg_id)

  def Clear(self):
    self.__init__()

  def Close(self):
    self.status = 'ready'

  def IsEmpty(self):
    return self.status == 'empty'

  def IsReady(self):
    return self.status == 'ready'




from g4camp.g4camp import g4camp
from time import time
import sys

n_events = 10
gun_args = { 'particle':   'e-', 
             'energy_GeV': 1000, 
             'position_m': [0,0,0],
             'direction':  [0,0,1]  }
optics = True

app = g4camp(primary_generator='gun', gun_args=gun_args, optics=optics)
app.setSkipMinMax(0.0001, 0.01)
app.setPhotonSuppressionFactor(100)
app.configure()

time0 = time()
for data in app.run(n_events):
    vertices = data.vertices
    tracks = data.tracks
    photon_cloud = data.photon_cloud
    print(f"{len(photon_cloud):>10}(x{app.ph_suppression_factor}) photons, {len(tracks):>6} tracks, {len(vertices):>6} vertices;")
print(f"# Run time:  {(time()-time0):.2f} sec, {((time()-time0)/n_events):.2f} sec/event")



from geant4_pybind import *

class EventAction(G4UserEventAction):

  def __init__(self, app):
    super().__init__()
    self.app = app
    self.data_buffer = app.data_buffer


  def BeginOfEventAction(self, evt):
    self.app.ph_counter = 0
    self.data_buffer.Clear()

  def EndOfEventAction(self, evt):
    self.data_buffer.CutEmptyTrackPoints()
    self.data_buffer.SortTrackPoints()
    self.data_buffer.Close()



from geant4_pybind import *
from g4camp.g4camp import g4camp
import sys

app = g4camp(optics=False, primary_generator='gun')
ui = G4UIExecutive(len(sys.argv), sys.argv)
visManager = G4VisExecutive("Quiet")
visManager.Initialize()
app.configure() # should be after initialization of visual manager
app.applyGeant4Command("/control/execute",  ["vis.mac"])
ui.SessionStart()



import sys, logging
logformat='[%(name)12s ] %(levelname)8s: %(message)s'
logging.basicConfig(format=logformat)

import pkg_resources
available_packages = [pkg.key for pkg in pkg_resources.working_set]
if 'geant4-pybind' in available_packages:
    from geant4_pybind import *
else:
    sys.exit("Package 'geant4_pybind' is not available.")

from g4camp.DetectorConstruction import DetectorConstruction
from g4camp.CustomPhysicsList import CustomPhysicsList
from g4camp.ActionInitialization import ActionInitialization
from g4camp.DataBuffer import DataBuffer

class g4camp:

    def __init__(self, optics=True, primary_generator="gun", gun_args={},
                 multithread=False, thread_num=4):
        # By some reason multithread mode does not faster than serial mode
        # FIXME: number of threads can not be configured from here (why???)
        #
        self.log = logging.getLogger('g4camp')
        #
        super().__init__()
        self.optics = optics                        # switch on/off optic physics
        self.primary_generator = primary_generator  # can be 'gun' or 'gps'
        self.multithread=multithread
        self.thread_num=thread_num
        #
        self.data_buffer = DataBuffer()
        self.gps_macro = ""
        #
        self.ph_suppression_factor = 10 # must be integer
        self.skip_mode = 'fraction'     # 'fraction' (relative to E_init) or 'GeV' (absolute)
        self.skip_min = 1.e-3           # min energy to skip particle (wrt. E_init or in GeV)
        self.skip_max = 1.e-1           # max energy to skip particle (wrt. E_init or in GeV)
        self.random_seed = 1
        # the following varialbes are to be set later
        self.E_skip_min = None          # min particle energy to skip
        self.E_skip_max = None          # max particle energy to skip
        self.E_init = None              # initial kinetic energy in GeV
        #
        self.ph_counter = None
        #
        if multithread:
            self.runManager = G4RunManagerFactory.CreateRunManager(G4RunManagerType.MT, thread_num)
        else:
            self.runManager = G4RunManagerFactory.CreateRunManager(G4RunManagerType.Serial)
        self.detConstruction = DetectorConstruction()
        self.physList = CustomPhysicsList(optics=self.optics)
        self.actInit = ActionInitialization(self, primary_generator=self.primary_generator, 
                                                  gun_args=gun_args)
        self.runManager.SetUserInitialization(self.detConstruction)
        self.runManager.SetUserInitialization(self.physList)
        self.runManager.SetUserInitialization(self.actInit)
        #
        self.setDefaultCuts()


    def configure(self):
        #self.actInit.stackingAction.SetPhotonSuppressionFactor(1./self.ph_fraction)
        self.setVerbose()
        if self.multithread: 
            UImanager.ApplyCommand(f"/run/numberOfThreads", [self.thread_num])
        if self.optics:
            self.configureOptics()
        self.runManager.Initialize()
        self.applyGeant4Command("/run/particle/dumpCutValues")
        if self.primary_generator == 'gps':
            self.applyGeant4Command("/control/execute", [self.gps_macro])
        G4Random.setTheSeed(self.random_seed)

    def setVerbose(self, control_verbose=0, tracking_verbose=0, run_verbose=0, em_process_verbose=0):
        self.applyGeant4Command("/control/verbose", [control_verbose])
        self.applyGeant4Command("/tracking/verbose", [tracking_verbose])
        self.applyGeant4Command("/run/verbose", [run_verbose])
        self.applyGeant4Command("/process/em/verbose", [em_process_verbose])

    def setCut(self, particle, cut_value, cut_unit):
        # to be invoked before 'configure()'
        self.applyGeant4Command("/run/setCutForAGivenParticle", ["e+", cut_value, cut_unit])

    def setDefaultCuts(self):
        # default cuts ensure all particle capable to produce Cherenkov light are tracked
        # i.e. E_threshold is about 260 keV (kinetic energy)
        self.setCut("e+", 0.055, "cm")
        self.setCut("e-", 0.055, "cm")
        self.setCut("gamm", 95., "cm")
        #self.applyGeant4Command("/cuts/setLowEdge", [760, "keV"]) # This can only lower energy limit

    def applyGeant4Command(self, command, arguments=[]):
        arg_string = ""
        for arg in arguments:
            arg_string += f" {arg}"
        UImanager = G4UImanager.GetUIpointer()
        UImanager.ApplyCommand(command + arg_string)

    def configureOptics(self):
        if not self.optics:
            self.log.warning("Optics is disabled")
            return
        self.applyGeant4Command("/process/optical/processActivation", ['Cerenkov', True])
        self.applyGeant4Command("/process/optical/processActivation", ['OpAbsorption', False])
        self.applyGeant4Command("/process/optical/processActivation", ['OpRayleigh', False])
        self.applyGeant4Command("/process/optical/processActivation", ['OpMieHG', False])
        self.applyGeant4Command("/process/optical/cerenkov/setStackPhotons", [True])

    def setGunParticle(self, pname):
        self.applyGeant4Command(f"/gun/particle {pname}")

    def setGunEnergy(self, ene_val, ene_unit):
        self.applyGeant4Command(f"/gun/energy {ene_val} {ene_unit}")

    def setGunPosition(self, x_val, y_val, z_val, pos_unit):
        self.applyGeant4Command(f"/gun/position {x_val} {y_val} {z_val} {pos_unit}")

    def setGunDirection(self, dx, dy, dz):
        self.applyGeant4Command(f"/gun/direction {dx} {dy} {dz}")

    def setGPSMacro(self, macro):
        if self.primary_generator != 'gps':
            self.log.warning("'primary_generator' was set to '{self.primary_generator}', switching it to 'gps'")
            self.primary_generator = 'gps'
        self.gps_macro = macro

    def setSkipMinMax(self, skip_min, skip_max):
        self.skip_min = skip_min
        self.skip_max = skip_max

    def setRandomSeed(self, val):
        self.random_seed = int(val)

    def setPhotonSuppressionFactor(self, val):
        self.ph_suppression_factor = float(val)

    def run(self, n_events):
        self.log.debug(f"random seed: {G4Random.getTheSeed()}")
        for i in range(n_events):
            self.runManager.BeamOn(1)
            yield self.data_buffer



from geant4_pybind import *

class RunAction(G4UserRunAction):

  def __init__(self, isMaster):
    super().__init__()
#    G4RunManager.GetRunManager().SetPrintProgress(1000)
    self.isMaster = isMaster
    self.printTiming = False
    if self.isMaster: 
      self.timer = G4Timer()

  def BeginOfRunAction(self, aRun):
    # inform the runManager to save random number seed
    G4RunManager.GetRunManager().SetRandomNumberStore(False)
    if self.isMaster:
      self.timer.Start()

  def SetPrintTiming(self, print_timing):
    self.printTiming = print_timing

  def EndOfRunAction(self, aRun):
    if self.isMaster:
      self.timer.Stop()
      if self.printTiming:
        import datetime
        dt_real = datetime.timedelta(seconds = self.timer.GetRealElapsed())
        dt_user = datetime.timedelta(seconds = self.timer.GetUserElapsed())
        dt_syst = datetime.timedelta(seconds = self.timer.GetSystemElapsed())
        print(f" Number of events: {aRun.GetNumberOfEvent()}")
        print(f" Real time:    {dt_real}")
        print(f" User time:    {dt_user}")
        print(f" System time:  {dt_syst}")



from geant4_pybind import *
import numpy as np

class StackingAction(G4UserStackingAction):

  def __init__(self, app):
    super().__init__()
    self.app = app
    self.data_buffer = app.data_buffer
    self.E_init = app.E_init
    self.ph_suppression_factor = app.ph_suppression_factor
    # Cherenkov threshold
    n = 1.35
    self.cerenkov_threshold_factor = 1./(1.-n**2)
    self.cerenkov_threshold_e = 511*keV * self.cerenkov_threshold_factor
    # E_th = mc^2 / sqrt(1-1/n^2),
    # where n = 1.35 for water, so 1/sqrt(1-1/n^2) = 1.4885
    #   for e+/e- E_th = 760 keV
    #   for mu    E_th = 157 MeV

  def ClassifyNewTrack(self, track):
    # If /process/optical/cerenkov/setTrackSecondariesFirst is set to True
    # then tracks are stacked even in the middle of their life,
    # so this method is invoked not only on particle creation.
    # This is temporary used to skip particles based on their energy,
    # because the possibility to kill particle in Stepping action is not 
    # yet available (G4Track.SetTrackStatus() is not yet pythonized).
    uid = track.GetTrackID()
    parentuid = track.GetParentID()
    position = track.GetPosition()/m
    time = track.GetGlobalTime()/ns
    pdgid = track.GetDefinition().GetPDGEncoding()
    momentum = track.GetMomentum()/GeV
    Etot = track.GetTotalEnergy()/GeV
    Ekin = track.GetKineticEnergy()/GeV
    # 
    # = Primary particles =======================
    #
    if uid == 1 and track.GetCurrentStepNumber() == 0: # only initial event vertex
      self.ph_suppression_factor = self.app.ph_suppression_factor
      self.app.E_init = Ekin
      self.app.E_skip_min = self.app.skip_min
      self.app.E_skip_max = self.app.skip_max
      if self.app.skip_mode == "fraction":
        self.app.E_skip_min *= self.app.E_init
        self.app.E_skip_max *= self.app.E_init
      #self.data_buffer.AddVertex([ [position.x, position.y, position.z, time],
      #                             [], 
      #                             [[pdgid, momentum.x, momentum.y, momentum.z, Etot]] ] )
      self.data_buffer.AddTrackPoint(uid, pdgid, position.x, position.y, position.z, time, Etot)
      #
      particle_name = track.GetDefinition().GetParticleName()
      energy_sting = G4BestUnit(Ekin*GeV, "Energy")
      x,y,z = position.x, position.y, position.z
      px, py, pz = momentum.x, momentum.y, momentum.z
      pmag = np.sqrt(px**2+py**2+pz**2)
      px, py, pz = px/pmag, py/pmag, pz/pmag
      self.app.log.info(f"{particle_name}:  Ekin={energy_sting}  position=({x:.1f}, {y:.1f}, {z:.1f}) m\
  direction=({px:.1f}, {py:.1f}, {pz:.1f}) ")
      return G4ClassificationOfNewTrack.fUrgent
    #
    # = Optical photons =========================
    #
    # - Store info and kill
    #
    if track.GetDefinition() == G4OpticalPhoton.OpticalPhoton():
      if self.app.ph_counter%self.ph_suppression_factor == 0:
        position = track.GetPosition()/m
        direction = track.GetMomentumDirection()/GeV
        time = track.GetGlobalTime()/ns
        wavelength = 1.23984193*1.e-6/Etot # in nm
        ph = np.array([position.x, position.y, position.z,
                       time,
                       direction.x, direction.y, direction.z,
                       wavelength])
        self.data_buffer.AddPhoton(ph)
        #if parentuid != 1: print(f"{time:.4f} photon    parent uid: {parentuid}")
      self.app.ph_counter += 1
      return G4ClassificationOfNewTrack.fKill
    #
    #
    # = All particles except optical photons ====
    #
    # === kill if E < E_skip_min
    if Etot < self.app.E_skip_min:
      return G4ClassificationOfNewTrack.fKill
    #
    # === store vertex for e- and e+ if E_skip_min < Etot < E_skip_max
    elif Etot > self.app.E_skip_min and Etot <= self.app.E_skip_max and uid != 1 \
         and track.GetDefinition() in [G4Electron.Electron(), G4Positron.Positron()]:
      self.app.data_buffer.AddVertex([ [position.x, position.y, position.z, time],
                                       [], 
                                       [[pdgid, momentum.x, momentum.y, momentum.z, Etot]] ] )
      #print(f"{time:.4f}   = to kill =    uid: {uid},  parent uid: {parentuid}, {pdgid}, energy: {Ekin:.4f}")
      return G4ClassificationOfNewTrack.fKill
    #
    # === store tracks 
    else:
      self.data_buffer.AddTrackPoint(uid, pdgid, position.x, position.y, position.z, time, Etot)
      return G4ClassificationOfNewTrack.fUrgent




import numpy as np

class Track():

  def __init__(self, unique_id, parent_id, pdg_id):
    self.unique_id = unique_id
    self.parent_id = parent_id
    self.pdg_id = pdg_id
    self.points = []

  def __str__(self):
    return f"UID: {self.unique_id},  parent_ID: {self.parent_id},  PDGID: {self.pdg_id}\n" + \
           f"points:\n" + \
           f"  {self.points[0]}\n" + \
           f"  ... \n" + \
           f"  {self.points[-1]}\n"

  def __repr__(self):
    return self.__str__()

  def AddPoint(self, point):
    self.points.append(point)
    pass



class DataBuffer():

  def __init__(self):
    self.vertices = []
    #
    self.track_points_per_bunch = 100000
    self.track_point_bunches = 1
    self.point_num = 0
    self.tracks = np.zeros((self.track_points_per_bunch, 7), dtype=float) 
    # 'track' row:  uid, pdgid, x[m], y[m], z[m], t[ns], Etot[GeV]
    self.photon_cloud = []
    self.status = 'empty' # 'empty' -> 'in_progress' -> 'ready'

  def AddVertex(self, vertex):
    self.status = 'in_progress'
    self.vertices.append(vertex)

  def AddPhoton(self, photon):
    self.status = 'in_progress'
    self.photon_cloud.append(photon)

  def CreateTrack(self, uid, parent_id, pdg_id):
    pass
#    self.status = 'in_progress'
#    self.tracks[uid] = Track(uid, parent_id, pdg_id)

  def AddTrackPoint(self, uid, pdgid, x, y, z, t, Etot):
    self.tracks[self.point_num] = [uid, pdgid, x, y, z, t, Etot]
    self.point_num += 1
    if self.point_num % self.track_points_per_bunch == 0:
      self.tracks = np.concatenate((self.tracks, np.zeros((self.track_points_per_bunch, 7), dtype=int)))
      self.track_point_bunches += 1

  def CutEmptyTrackPoints(self):
    self.tracks = self.tracks[:self.point_num]

  def SortTrackPoints(self):
    self.tracks = self.tracks[np.lexsort((self.tracks[:,5], self.tracks[:,0]))]

  def Clear(self):
    self.__init__()

  def Close(self):
    self.status = 'ready'

  def IsEmpty(self):
    return self.status == 'empty'

  def IsReady(self):
    return self.status == 'ready'

